# Copyright (C) 2024, UChicago Argonne, LLC
# Licensed under the 3-clause BSD license.  See accompanying LICENSE.txt file
# in the top-level directory.

"""
Rules
only allowed to import pbs_util.py
no importing anything except basic python packages
no creating instances of classes globally, this could lead to memory leaks.
do not import pbs, have it passed in, this allows testing and mocking more easily.

Remember this is imported into pbs in the server and pbs_python.
Pattern for importing this:
    import PBS_Utils
    import PBS_Utils.pbs_hook_library
    importlib.reload(PBS_Utils)
    importlib.reload(PBS_Utils.pbs_hook_library)
    from PBS_Utils.pbs_hook_library import add_node_state_node_for_an_event

Never do this in a hook or hook code:
    obj = datetime.strptime(ts_str, "%Y-%m-%dT%H:%M:%S.%f")

TODO: figure this out
    1.  Don't directly log, raise exception with message and catch it on the hook side.
"""
from typing import Optional


class PBS_Serializer:
    def __init__(self, pbs):
        self.pbs = pbs
        self.server_object_key_skips = (
            "attributes",
            "_attributes_hook_set",
            "_attributes_unknown",
            "topology_info",  # contains hwloc
        )

    def serialize_pbs_module(self):
        pbs = self.pbs
        dct = {}
        for key in dir(pbs):
            val = getattr(pbs, key)
            if (
                not callable(val)
                and not key.startswith("__")
                and key != "EXPORTED_TYPES_DICT"
                and not repr(val).startswith("<module ")
            ):
                dct[key] = val
            else:
                pass
        return dct

    def serialize_server_object_value(self, val, include_none=True):
        pbs = self.pbs
        bt = self.pbs.v1._base_types
        et = self.pbs.v1._exc_types
        val_type = type(val)
        # if isinstance(val, (int, float, str, bool)):
        #    result = val  # DO NOT DO THIS!
        if isinstance(val, str):
            result = str(val)
        elif isinstance(val, int):
            result = int(val)
        elif isinstance(val, float):
            result = float(val)
        elif isinstance(val, bool):
            result = bool(val)
        elif isinstance(val, dict):
            # note, this is weird, it's dict, but will not work right
            # this fixes at least Variable_List
            val = dict(val)
            dct = {}
            for k, v in val.items():
                if v is not None or include_none:
                    new_val = self.serialize_server_object_value(v, include_none=include_none)
                    if new_val is not None or include_none:
                        dct[k] = new_val
            result = dct
        elif isinstance(val, pbs.v1._base_types.pbs_resource):
            result = dict(val)
        elif isinstance(val, (bt.pbs_bool,)):
            # _derived_types = (bool,)
            result = bool(val)
        elif isinstance(val, (bt.priority, bt.job_state)):
            # (_generic_attr, int) or (int, )
            result = int(val._value)
        elif isinstance(val, (bt.path, bt.sandbox, bt.version, bt.name, bt.project)):
            # (_generic_attr, str) or (str, _generic_attr)
            result = str(val)
        elif isinstance(
            val,
            (
                bt.join_path,
                bt.user_list,
                bt.group_list,
                bt.depend,
                bt.checkpoint,
                bt.exec_host,
                bt.keep_files,
                bt.mail_points,
                bt.staging_list,
                bt.range,
                bt.state_count,
                bt.place,
                bt.hold_types,
            ),
        ):
            # (_generic_attr,)
            result = str(val)
        # elif isinstance(val, pbs.v1._base_types._generic_attr):
        #     # note this will probably lose information
        #     result = str(val)
        elif str(val_type).startswith("<class 'pbs.v1._svr_types."):
            result = self.serialize_server_object(val, include_none=include_none)
        elif str(val_type).startswith("<class 'pbs.v1._base_types."):
            result = self.serialize_server_object(val, include_none=include_none)
        elif val_type in (tuple, list, set):
            lst = []
            for o in val:
                if o is not None or include_none:
                    lst.append(self.serialize_server_object_value(o, include_none=include_none))
            result = val_type(lst)
        elif val_type == dict:
            dct = {}
            for k, v in val.items():
                if v is not None or include_none:
                    new_val = self.serialize_server_object_value(v, include_none=include_none)
                    if new_val is not None or include_none:
                        dct[k] = new_val
            result = dct
        else:
            result = val
        if not include_none and result == "":
            result = None
        return result

    def serialize_server_object(self, obj, include_none=True):
        # include_node==False removes None values
        dct = {}
        for key in dir(obj):
            if not key.startswith("__") and key not in self.server_object_key_skips:
                val = getattr(obj, key)
                if not callable(val):
                    if key == "_param" and isinstance(val, dict):
                        for k, v in val.items():
                            if v is not None or include_none:
                                dct[k] = self.serialize_server_object_value(v, include_none=include_none)
                    else:
                        if val is not None or include_none:
                            new_val = self.serialize_server_object_value(val, include_none=include_none)
                            if new_val is not None or include_none:
                                dct[key] = new_val
        return dct


def add_node_state_via_server(server_obj, event_obj, context: str, node: str, state) -> Optional[str]:
    """for server side hooks or to offline any node."""
    err = None
    vnode_obj = None

    try:
        vnode_obj = server_obj.vnodes[node]
    except KeyError:
        err = f"{context}; Missing {node} in server_obj.vnodes"
    else:
        if event_obj and hasattr(event_obj, "vnode_list_fail"):
            event_obj.vnode_list_fail[node] = vnode_obj

    if vnode_obj:
        if vnode_obj.state:
            vnode_obj.state += state
        else:
            vnode_obj.state = state
        vnode_obj.comment = err
    return err


def add_node_state_node_for_an_event(event_obj, node: str, state, comment=None) -> Optional[str]:
    """for offing nodes for a mom associated with the event."""
    err = None
    vnode_obj = None

    try:
        vnode_obj = event_obj.vnode_list_fail[node]
    except KeyError:
        try:
            vnode_obj = event_obj.vnode_list[node]
        except KeyError:
            err = f"Missing {node} in event_obj.vnode_list."
        else:
            event_obj.vnode_list_fail[node] = vnode_obj

    if vnode_obj:
        if vnode_obj.state:
            vnode_obj.state += state
        else:
            vnode_obj.state = state
        if comment:
            vnode_obj.comment = comment
    return err
