#!/usr/bin/env python3

# Copyright (C) 2024, UChicago Argonne, LLC
# Licensed under the 3-clause BSD license.  See accompanying LICENSE.txt file
# in the top-level directory.

import json
import os
import uuid
import time
import socket
import inspect
import functools
from os import linesep
from contextlib import contextmanager
from typing import List, TypedDict
from PBS_Utils.pbs_util import get_now

OT_TRACER = None  # there must be only one OT tracer.


def formatter(span):
    return span.to_json(indent=None) + linesep


def get_callinfo(jump_back_count=1, raw=True):
    current_frame = inspect.currentframe()
    frame, filename, line_number, function_name, lines, index = inspect.getouterframes(current_frame)[jump_back_count]
    fpath, fname = os.path.split(filename)
    lineno = frame.f_lineno
    if raw:
        result = {"fname": fname, "lineno": lineno, "filename": filename, "index": index, "function_name": function_name}
    else:
        result = "%-13s:L#%s" % (fname, lineno)
    return result


# ---- vvv ---- TracerLogger ---- vvv ----
class TimeIt:
    def __init__(self):
        self.ts = None
        self.te = None

    def __enter__(self):
        self.ts = time.time()
        return self

    def __exit__(self, *args):
        self.te = time.time()

    def get_td(self):
        return self.te - self.ts


class Timings(TypedDict):
    name: str  # name of timing
    ts: float  # time start of scope
    te: float  # time end of scope
    call_i: dict  # call in info
    call_e: dict  # call exit info
    td_src: float  # time of just the source code
    td_total: float  # total time


class TracerSimple:
    def __init__(self, name, output_path):
        self.name = name
        self.current_span_name = "instrument_setup"
        self.output_path = output_path
        self.traces: List[Timings] = []
        self.instrument_timings = []
        self.ts_inst = time.time()
        with TimeIt() as ti:
            self.call_i = get_callinfo(jump_back_count=5)
        self.instrument_timings.append(ti.get_td())
        self.te_inst = self.ts_inst
        with TimeIt() as ti:
            self.call_e = get_callinfo(jump_back_count=5)
        self.instrument_timings.append(ti.get_td())
        self.td_code = 0.0
        self.ts_code = time.time()
        self.te_code = self.ts_code

    @contextmanager
    def start_as_current_span(self, name, *args, **kwargs):
        with TimeIt() as ti_enter:
            call_i = get_callinfo(jump_back_count=3)
        td_start_inst = ti_enter.get_td()
        self.instrument_timings.append(td_start_inst)
        self.current_span_name = name
        with TimeIt() as ti_src:
            yield self
        ts_src = ti_src.ts
        te_src = ti_src.te
        with TimeIt() as ti_exit:
            call_e = get_callinfo(jump_back_count=3)
        td_end_inst = ti_exit.get_td()
        self.instrument_timings.append(td_end_inst)
        td_src = te_src - ts_src
        td_total = td_start_inst + td_end_inst + td_src
        self.traces.append(
            Timings(name=name, ts=ts_src, te=te_src, call_i=call_i, call_e=call_e, td_src=td_src, td_total=td_total)
        )

    def shutdown(self):
        self.te_code = time.time()
        self.td_code = self.te_code - self.ts_code
        self.current_span_name = "shutdown"
        with TimeIt() as ti:
            self.call_e = get_callinfo(jump_back_count=4)
        self.instrument_timings.append(ti.get_td())
        self.te_inst = time.time()
        td_src = self.te_code - self.ts_code
        td_total = self.te_inst - self.ts_inst
        self.traces.append(
            Timings(
                name=self.name,
                ts=self.ts_code,
                te=self.te_code,
                call_i=self.call_i,
                call_e=self.call_e,
                td_src=td_src,
                td_total=td_total,
            )
        )
        if self.output_path:
            with open(self.output_path, "w") as fd:
                for dct in self.traces:
                    fd.write(json.dumps(dct))
                    fd.write("\n")

    def get_current_span(self):
        # todo: this would require a refactor to create new spans
        return self


# ---- ^^^ ---- TracerLogger ---- ^^^ ----


# ---- vvv ---- TracerFake ---- vvv ----
class TracerFake:
    def __init__(self, name, output_path):
        self.name = name
        self.output_path = output_path

    @contextmanager
    def start_as_current_span(self, name, *args, **kwargs):
        yield self

    def shutdown(self):
        pass

    def get_current_span(self):
        return self


# ---- ^^^ ---- TracerFake ---- ^^^ ----


def get_tracer(trace_name, output_path, tracer_class):
    return tracer_class(trace_name, output_path)


@contextmanager
def instrument_setup(tracer_name, output_path, tracer_class=None):
    if tracer_name and not tracer_class:
        global OT_TRACER
        from opentelemetry import trace
        from opentelemetry.sdk.trace import TracerProvider, Tracer
        from opentelemetry.sdk.trace.export import BatchSpanProcessor, ConsoleSpanExporter

        # Only can have one tracer per program.  Do not call more than once per program.
        outer_tracer: Tracer = trace.get_tracer(tracer_name)
        OT_TRACER = outer_tracer
        # trace.set_tracer_provider(TracerProvider(shutdown_on_exit=True))
        trace.set_tracer_provider(TracerProvider(shutdown_on_exit=False))
        fd = open(output_path, "a")
        bsp = BatchSpanProcessor(ConsoleSpanExporter(out=fd, formatter=formatter))
        trace.get_tracer_provider().add_span_processor(bsp)
        try:
            yield outer_tracer
        finally:
            fd.flush()
            bsp.shutdown()
            fd.close()
            OT_TRACER = None
    else:
        tracer_class = tracer_class if tracer_class else TracerFake
        outer_tracer = get_tracer(tracer_name, output_path, tracer_class)
        try:
            yield outer_tracer
        finally:
            outer_tracer.shutdown()


def ot_decorate_span(tracer_inner):
    """wrap a function call with opentelemetry."""
    pid = os.getpid()
    hostname = socket.gethostname()

    def outer_decorate(func):

        @functools.wraps(func)
        def decorate(*args, **kwargs):
            global OT_TRACER
            if OT_TRACER:
                from opentelemetry.trace.status import Status, StatusCode

                current_frame = inspect.currentframe()
                frame, filename, line_number, function_name, lines, index = inspect.getouterframes(current_frame)[1]
                now_str = get_now().strftime("%Y-%m-%d %H:%M:%S.%f")
                # all long calls must be done before the with line.
                with tracer_inner.start_as_current_span(f"{func.__name__}") as span:
                    span.set_attribute("hostname", hostname)
                    span.set_attribute("pid", pid)
                    span.set_attribute("timestamp", now_str)
                    span.set_attribute("filename", filename)
                    span.set_attribute("lineno", line_number)
                    span.set_attribute("func_name", function_name)
                    try:
                        result = func(*args, **kwargs)
                    except Exception as err:
                        span.record_exception(err)
                        span.set_status(Status(StatusCode.ERROR))
                        raise
            else:
                # we don't have a tracer anymore, but let's still execute.
                result = func(*args, **kwargs)
            return result

        return decorate

    return outer_decorate
