#!/usr/bin/env python3

# Copyright (C) 2024, UChicago Argonne, LLC
# Licensed under the 3-clause BSD license.  See accompanying LICENSE.txt file
# in the top-level directory.

import tempfile
import time
import abc
import json
import os
import socket
import subprocess
import uuid
from pathlib import Path
from pprint import pformat
from typing import Dict, List, Optional, Union
from datetime import datetime
from typing import TypedDict
from PBS_Utils.pbs_util import timedelta_str_to_td
from PBS_Utils.pbs_util import get_now


class PBSMetaData(TypedDict):
    timestamp: datetime
    pbs_version: str
    pbs_server: str
    hostname: str


class PBSResponseData(TypedDict):
    metadata: PBSMetaData
    data: Optional[Union[Dict, List]]


class PBSResvData(TypedDict):
    name: str


class PBSNodeData(TypedDict):
    Mom: str
    Port: str
    last_state_change_time: str  # 1612825851
    last_used_time: str  # 1612825851
    license: str
    ntype: str
    pbs_version: str
    pcpus: str  # 8
    resources_assigned: Dict[str, str]
    resources_available: Dict[str, str]
    resv_enable: str  # True
    server_instance_id: str
    sharing: str
    state: str


def decorator_reconnect(method):
    def decorate(slf, *args, **kwargs):
        now = time.time()
        if (now - slf.last_connect) > slf.timeout:
            if slf.debug:
                print(f"reconnecting due to timeout of {slf.timeout}. last_connect:{slf.last_connect} now:{now}")
            slf.reconnect()
        result = method(slf, *args, **kwargs)
        return result

    return decorate


class PBSIFLError(Exception):
    pass


class PBSInterfaceABC(metaclass=abc.ABCMeta):
    def __init__(self, machine_name=None, debug=False):
        self.machine_name = machine_name
        self.debug = debug

    @abc.abstractmethod
    def get_pbs_nodes(self, nodes=[]) -> PBSResponseData:
        raise NotImplemented()

    @abc.abstractmethod
    def get_pbs_vnodes(self, nodes=[]) -> PBSResponseData:
        raise NotImplemented()

    @abc.abstractmethod
    def get_qstat(self):
        raise NotImplemented()


def get_metadata(machine_name) -> PBSMetaData:
    now = get_now(notz=True)
    version = None
    # FIXME: if you do this, it will destroy the connection!
    server = ""  # pbs.server().name
    result_dct = {
        "timestamp": now,
        "pbs_version": version,
        "pbs_server": server,
        "hostname": socket.gethostname(),
        "machine_name": machine_name,
        "pid": os.getpid(),
        "uuid": uuid.uuid4().hex,
    }
    return result_dct


class PBSCommandInterface(PBSInterfaceABC):
    def __init__(self, machine_name=None, debug=False):
        self.cmd_pbsnodes = "/opt/pbs/bin/pbsnodes"
        self.cmd_qstat = "/opt/pbs/bin/qstat"

        cmd_lst = [self.cmd_pbsnodes, self.cmd_qstat]
        missing_cmd_lst = []
        for cmd in cmd_lst:
            if not os.path.exists(cmd):
                missing_cmd_lst.append(cmd)
        if len(missing_cmd_lst) > 0:
            raise ImportError(f"cannot load commands:{','.join(missing_cmd_lst)}")

        super().__init__(machine_name=machine_name, debug=debug)

    def get_pbs_nodes(self, nodes=[]) -> PBSResponseData:
        # FIXME: get nodes to work
        cmd_lst = [self.cmd_pbsnodes, "-aF", "json"]
        pipe = subprocess.PIPE
        proc = subprocess.Popen(cmd_lst, stdout=pipe, stderr=pipe)
        stdout: bytes
        stderr: bytes
        stdout, stderr = proc.communicate()
        returncode = proc.returncode
        if returncode == 0:
            node_data = json.loads(stdout.decode().replace("\n", ""))
        else:
            if b"no node list" in stderr:
                node_data = {
                    "nodes": {},
                    "timestamp": time.time(),
                    # 'pbs_version': ?,
                    # 'pbs_server': ?,
                }
            else:
                node_data = {}
            if self.debug:
                print(f"Error: cmd_lst: {cmd_lst} returned {returncode}")
                print(f"stdout:{stdout}")
                print(f"stderr:{stderr}")
        result_dct = {
            "metadata": get_metadata(self.machine_name),
            "data": node_data,
        }
        return result_dct

    def pbs_qstat(self):
        cmd_lst = [self.cmd_qstat, "-t", "-f", "-F", "json"]
        pipe = subprocess.PIPE
        proc = subprocess.Popen(cmd_lst, stdout=pipe, stderr=pipe)
        stdout: bytes
        stderr: bytes
        stdout, stderr = proc.communicate()
        returncode = proc.returncode
        if self.debug:
            print(f"Error: cmd_lst: {cmd_lst} returned {returncode}")
            print(f"stdout:{stdout}")
            print(f"stderr:{stderr}")
        if returncode == 0:
            dct = json.loads(stdout.decode().replace("\n", ""))
        else:
            dct = None
        return dct

    def get_pbs_vnodes(self, nodes=[]) -> PBSResponseData:
        # FIXME: get nodes to work
        cmd_lst = [self.cmd_pbsnodes, "-avF", "json"]
        pipe = subprocess.PIPE
        proc = subprocess.Popen(cmd_lst, stdout=pipe, stderr=pipe)
        stdout: bytes
        stderr: bytes
        stdout, stderr = proc.communicate()
        returncode = proc.returncode
        if returncode == 0:
            node_data = json.loads(stdout.decode().replace("\n", ""))
        else:
            if b"no node list" in stderr:
                node_data = {
                    "nodes": {},
                    "timestamp": time.time(),
                    # 'pbs_version': ?,
                    # 'pbs_server': ?,
                }
            else:
                node_data = {}
            if self.debug:
                print(f"Error: cmd_lst: {cmd_lst} returned {returncode}")
                print(f"stdout:{stdout}")
                print(f"stderr:{stderr}")
        result_dct = {
            "metadata": get_metadata(self.machine_name),
            "data": node_data,
        }
        return result_dct


class PBSIFLInterface(PBSInterfaceABC):

    def __init__(self, machine_name=None, host=None, debug=False, pbs_ifl=None):
        if pbs_ifl is None:
            import pbs_ifl
        self.pbs_ifl = pbs_ifl
        self.host = host
        self.last_connect = time.time()
        self.conn = pbs_ifl.pbs_connect(None)
        self.timeout = 1800 - 60
        if self.conn == -1:
            raise PBSIFLError(f"Error: Could not connect to pbs_ifl on host:{self.host}.")
        super().__init__(machine_name, debug=debug)

    def __del__(self):
        if hasattr(self, "conn") and self.conn:
            self.pbs_ifl.pbs_disconnect(self.conn)

    def disconnect(self):
        if hasattr(self, "conn") and self.conn:
            self.pbs_ifl.pbs_disconnect(self.conn)

    def reconnect(self):
        now = time.time()
        if self.debug:
            print(f"reconnecting to pbs_ifl.  last_connect:{self.last_connect} now:{now}")
        self.disconnect()
        self.conn = self.pbs_ifl.pbs_connect(None)
        self.last_connect = now
        if self.conn == -1:
            if self.host is None:
                host = self.get_pbs_server_host()
            else:
                host = self.host
            raise PBSIFLError(f"Error: Could not connect to pbs_ifl on host:{host}.")

    def get_pbs_server_host(self):
        server = self.pbs_ifl.get_pbs_server()
        return server

    def check_for_error(self, func, bstat_head):
        errmsg = self.pbs_ifl.pbs_geterrmsg(self.conn)
        errno = self.pbs_ifl.get_pbs_errno()
        if self.debug:
            print(f"pbs_ifl errno:{errno}, func:{func}, bstat_head:{bstat_head}")
        if errno == 15064:
            pass  # Server has no node list
        elif errno != 0:
            # errors we want to comment on, but still throw an error
            if errno == 15136:
                print(f"pbs_ifl hook error errno:{errno} errmsg:{errmsg} func:{func} bstat_head:{bstat_head}")
            elif errno == 15020 or errno == 15031:
                self.reconnect()
                raise PBSIFLError(pformat(f"pbs_ifl(reconnected) errno:{errno} txt:{self.pbs_ifl.pbse_to_txt(errno)}"))
            raise PBSIFLError(pformat(f"pbs_ifl errno:{errno} txt:{self.pbs_ifl.pbse_to_txt(errno)}"))

    def get_manager_cmd(self, command):
        # FIXME: this should be in pbs_ifl as a reverse like in _svr_types.py
        pbs_ifl = self.pbs_ifl
        if command == "create":
            result = pbs_ifl.MGR_CMD_CREATE
        elif command == "delete":
            result = pbs_ifl.MGR_CMD_DELETE
        elif command == "set":
            result = pbs_ifl.MGR_CMD_SET
        elif command == "unset":
            result = pbs_ifl.MGR_CMD_UNSET
        elif command == "list":
            result = pbs_ifl.MGR_CMD_LIST
        elif command == "print":
            result = pbs_ifl.MGR_CMD_PRINT
        elif command == "active":
            result = pbs_ifl.MGR_CMD_ACTIVE
        elif command == "import":
            result = pbs_ifl.MGR_CMD_IMPORT
        elif command == "export":
            result = pbs_ifl.MGR_CMD_EXPORT
        else:
            raise Exception(f"invalid command {command}")
        return result

    def get_manager_obj(self, obj):
        # FIXME: this should be in pbs_ifl as a reverse like in _svr_types.py
        pbs_ifl = self.pbs_ifl
        if obj == "server":
            result = pbs_ifl.MGR_OBJ_SERVER
        elif obj == "queue":
            result = pbs_ifl.MGR_OBJ_QUEUE
        elif obj == "job":
            result = pbs_ifl.MGR_OBJ_JOB
        elif obj == "node":
            result = pbs_ifl.MGR_OBJ_NODE
        elif obj == "reservation":
            result = pbs_ifl.MGR_OBJ_RESV
        elif obj == "resource":
            result = pbs_ifl.MGR_OBJ_RSC
        elif obj == "host":
            result = pbs_ifl.MGR_OBJ_HOST
        elif obj == "hook":
            result = pbs_ifl.MGR_OBJ_HOOK
        elif obj == "pbs_hook":
            result = pbs_ifl.MGR_OBJ_PBS_HOOK
        elif obj == "jobarray_parent":
            result = pbs_ifl.MGR_OBJ_JOBARRAY_PARENT
        elif obj == "subjob":
            result = pbs_ifl.MGR_OBJ_SUBJOB
        else:
            raise Exception(f"invalid object {obj}")
        return result

    @decorator_reconnect
    def _get_pbs_job_data(self, jobid: str, attrib, extend, err_handler) -> Dict:
        """
        return a dictionary with jobid as the key and the entire job data as the value.
        """
        pbs_ifl = self.pbs_ifl
        bstat_head = pbs_ifl.pbs_statjob(self.conn, jobid, attrib, extend)
        try:
            self.check_for_error(pbs_ifl.pbs_statjob, bstat_head)
        except PBSIFLError as err:
            if err_handler:
                err_handler(err)
            else:
                raise
        data = process_standard_bstat_head(bstat_head, pbs_ifl)
        return data

    def get_pbs_jobs(self, jobids=[], attrib=None, extend=None, score=False, err_handler=None) -> PBSResponseData:
        if score:
            try:
                # FIXME: This is not the dotted server.  How do we get the full server.
                # server = self.get_pbs_server()
                # job_sort_formula = self.get_pbs_server(attrl=attrl)['data'][server]['job_sort_formula']
                attrl = self.get_attrl("job_sort_formula")
                dct = list(self.get_pbs_server(attrl=attrl)["data"].values()).pop()
                job_sort_formula = dct["job_sort_formula"]
            except Exception:
                job_sort_formula = None
        else:
            job_sort_formula = None
        dct = {}
        if not jobids:
            job_dct = self._get_pbs_job_data("", attrib, extend, err_handler)
            dct.update(job_dct)
        else:
            for jobid in jobids:
                job_dct = self._get_pbs_job_data(jobid, attrib, extend, err_handler)
                dct.update(job_dct)

        if score and job_sort_formula:
            for jobid, job_dct in dct.items():
                score_vars = {"walltime": 0}
                # set server eligible_time_enable = True
                score_vars["eligible_time"] = timedelta_str_to_td(job_dct["eligible_time"]).total_seconds()
                for key, value in job_dct["Resource_List"].items():
                    if key in ["walltime"]:
                        try:
                            score_vars[key] = timedelta_str_to_td(value).total_seconds()
                        except (TypeError, ValueError):
                            pass
                    else:
                        try:
                            score_vars[key] = int(value)
                        except (TypeError, ValueError):
                            pass
                if self.debug:
                    print(pformat(job_sort_formula))
                    print(pformat(score_vars))
                score_value = eval(job_sort_formula, score_vars)
                job_dct["score"] = score_value

        result_dct = {
            "metadata": get_metadata(self.machine_name),
            "data": dct,
        }
        return result_dct

    @decorator_reconnect
    def _get_pbs_node_data(self, node: str) -> Dict[str, PBSNodeData]:
        pbs_ifl = self.pbs_ifl
        bstat_head = pbs_ifl.pbs_statnode(self.conn, node, None, None)
        self.check_for_error(pbs_ifl.pbs_statnode, bstat_head)
        data = process_standard_bstat_head(bstat_head, pbs_ifl)
        return data

    def get_pbs_nodes(self, nodes=[], debug=False) -> PBSResponseData:
        data = {}
        if not nodes:
            dct = self._get_pbs_node_data("")
            data.update(dct)
        else:
            for node in nodes:
                dct = self._get_pbs_node_data(node)
                data.update(dct)
        if self.debug:
            print(pformat(data))
        result_dct = {
            "metadata": get_metadata(self.machine_name),
            "data": data,
        }
        return result_dct

    def get_qstat(self):
        raise NotImplemented()

    @decorator_reconnect
    def _get_pbs_vnode_data(self, node: str) -> Dict[str, PBSNodeData]:
        pbs_ifl = self.pbs_ifl
        func = pbs_ifl.pbs_statvnode
        bstat_head = func(self.conn, node, None, None)
        self.check_for_error(func, bstat_head)
        data = process_standard_bstat_head(bstat_head, pbs_ifl)
        return data

    def get_pbs_vnodes(self, nodes=[]) -> PBSResponseData:
        node_data = {}
        if not nodes:
            node_dct = self._get_pbs_vnode_data("")
            node_data.update(node_dct)
        else:
            for node in nodes:
                node_dct = self._get_pbs_vnode_data(node)
                node_data.update(node_dct)
        if self.debug:
            print(pformat(node_data))
        result_dct = {
            "metadata": get_metadata(self.machine_name),
            "data": node_data,
        }
        return result_dct

    def get_attropl(self, name, value=None, resource=None, next_attropl=None, batch_op: int = None):
        pbs_ifl = self.pbs_ifl
        attropl = pbs_ifl.attropl()
        attropl.name = name
        attropl.value = value
        if resource:
            attropl.resource = resource
        if next_attropl:
            attropl.next = next_attropl
        if batch_op:
            attropl.op = batch_op
        # todo: where is freeattropl or free_attropl hiding...
        return attropl

    def get_attrl(self, name, value=None, resource=None, next_attrl=None):
        pbs_ifl = self.pbs_ifl
        attrl = pbs_ifl.attrl()
        attrl.name = name
        attrl.value = value
        if resource:
            attrl.resource = resource
        if next_attrl:
            attrl.next = next_attrl
        # todo: where is free_attrl_list or free_attrl
        return attrl

    @decorator_reconnect
    def _get_pbs_server_data(self, attrl=None) -> Dict[str, PBSNodeData]:
        pbs_ifl = self.pbs_ifl
        func = pbs_ifl.pbs_statserver
        bstat_head = func(self.conn, attrl, None)
        self.check_for_error(func, bstat_head)
        data = process_standard_bstat_head(bstat_head, pbs_ifl)
        return data

    def get_pbs_server(self, attrl=None) -> PBSResponseData:
        data = self._get_pbs_server_data(attrl=attrl)
        if self.debug:
            print(pformat(data))
        result_dct = {
            "metadata": get_metadata(self.machine_name),
            "data": data,
        }
        return result_dct

    @decorator_reconnect
    def _get_pbs_queue_data(self, queue_name=None) -> Dict[str, PBSNodeData]:
        pbs_ifl = self.pbs_ifl
        func = pbs_ifl.pbs_statque
        bstat_head = func(self.conn, queue_name, None, None)
        self.check_for_error(func, bstat_head)
        data = process_standard_bstat_head(bstat_head, pbs_ifl)
        return data

    def get_pbs_queues(self, queue_name=None) -> PBSResponseData:
        data = self._get_pbs_queue_data(queue_name=queue_name)
        if self.debug:
            print(pformat(data))
        result_dct = {
            "metadata": get_metadata(self.machine_name),
            "data": data,
        }
        return result_dct

    @decorator_reconnect
    def _get_pbs_resv_data(self, resvid: str) -> Dict[str, PBSResvData]:
        pbs_ifl = self.pbs_ifl
        bstat_head = pbs_ifl.pbs_statresv(self.conn, resvid, None, None)
        self.check_for_error(pbs_ifl.pbs_statresv, bstat_head)
        data = process_standard_bstat_head(bstat_head, pbs_ifl)
        return data

    def get_pbs_reservations(self, resvids=[]) -> PBSResponseData:
        dct = {}
        if not resvids:
            resv_dct = self._get_pbs_resv_data("")
            dct.update(resv_dct)
        else:
            for resvid in resvids:
                resv_dct = self._get_pbs_job_data(resvid)
                dct.update(resv_dct)
        result_dct = {
            "metadata": get_metadata(self.machine_name),
            "data": dct,
        }
        return result_dct

    def pbs_stathook(self, hook_name, attrib_lst, extend) -> PBSResponseData:
        pbs_ifl = self.pbs_ifl
        bstat_head = pbs_ifl.pbs_stathook(self.conn, hook_name, attrib_lst, extend)
        self.check_for_error(pbs_ifl.pbs_stathook, bstat_head)
        data = process_standard_bstat_head(bstat_head, pbs_ifl)
        return data

    def pbs_statserver(self, pbs_server, attrl_name):
        pbs_ifl = self.pbs_ifl
        extend = None
        if attrl_name:
            attrl = pbs_ifl.attrl()
            attrl.name = attrl_name
        else:
            attrl = ""
        bstat_head = pbs_ifl.pbs_statserver(self.conn, attrl, extend)
        self.check_for_error(pbs_ifl.pbs_manager, bstat_head)
        data = process_standard_bstat_head(bstat_head, pbs_ifl)
        if attrl_name:
            server_data = data[pbs_server]
            try:
                result = server_data[attrl_name]
            except KeyError:
                result = None  # an attribute is not set in the server.
        else:
            result = data
        return result

    def pbs_manager(self, command, object_type, object_name="", attropl=None, extend=None) -> PBSResponseData:
        if type(command) == str:
            command = self.get_manager_cmd(command)
        if type(object_type) == str:
            object_type = self.get_manager_obj(object_type)
        pbs_ifl = self.pbs_ifl
        errno = pbs_ifl.pbs_manager(self.conn, command, object_type, object_name, attropl, extend)
        self.check_for_error(pbs_ifl.pbs_manager, None)
        return errno

    def pbs_manager_set_and_wait(self, attrl_name, attropl_value, delay=0.1):
        command = self.pbs_ifl.MGR_CMD_SET
        obj_type = self.pbs_ifl.MGR_OBJ_SERVER
        pbs_server = self.get_pbs_server()
        attropl = self.get_attropl(attrl_name, attropl_value)
        self.pbs_manager(command, obj_type, attropl=attropl)
        count = 0
        max_count = 32
        while True:
            data = self.pbs_statserver(pbs_server, attrl_name)
            if data.lower() == attropl_value.lower():
                break
            time.sleep(delay)
            count += 1
            if count >= max_count:
                raise Exception(f"error setting value, check count exceeded:{count}")

    def pbs_submit(self, job_script, attropl_lst):
        pbs_ifl = self.pbs_ifl
        prev_attropl = None
        attropl = pbs_ifl.attropl()
        attropl_first = attropl
        for attropl_dct in attropl_lst:
            attropl.name = attropl_dct["name"]
            attropl.value = attropl_dct["value"]
            if "resource" in attropl_dct:
                attropl.resource = attropl_dct["resource"]
            if prev_attropl:
                prev_attropl.next = attropl
            prev_attropl = attropl
            attropl = pbs_ifl.attropl()
        data = pbs_ifl.pbs_submit(self.conn, attropl_first, job_script, "", "")
        # todo: figure out how to check for errors and not mess up the return.
        # self.check_for_error(pbs_ifl.pbs_submit, data)
        return data

    def pbs_delete(self, jobid, wait=False, delay=0.1):
        pbs_ifl = self.pbs_ifl
        status = pbs_ifl.pbs_deljob(self.conn, jobid, None)
        if status == 0:
            if wait:
                count = 0
                max_count = 32

                def handler(err):
                    errno = self.pbs_ifl.get_pbs_errno()
                    if errno == 15001:
                        pass
                    elif errno == 15139 and str(err) == "'pbs_ifl errno:15139 txt:Job has finished'":
                        pass
                    else:
                        raise (err)

                while True:
                    data = self._get_pbs_job_data(jobid, None, None, handler)
                    if jobid not in data:
                        break
                    time.sleep(delay)
                    count += 1
                    if count >= max_count:
                        raise Exception(f"Error waiting for job {jobid} to be deleted, check count exceeded:{count}")
            else:
                pass  # we are not waiting, just returning, we already have a status of 0.
        else:
            raise Exception(f"Job {jobid} was not successfully deleted:{status}")


def process_standard_bstat_head(bstat_head, pbs_ifl):
    data = {}
    bstat_root = bstat_head
    while bstat_head:
        dct = {}
        if bstat_head:
            attrl = bstat_head.attribs
            while attrl:
                key = attrl.name
                resource = attrl.resource
                value = attrl.value
                if not resource:
                    if key in dct:
                        raise Exception(f"key:{key} is already set and resource didn't exist on the previous loop.")
                    dct[key] = value
                else:
                    if key not in dct:
                        dct[key] = {}
                    dct[key].update(
                        {
                            resource: value,
                        }
                    )
                    # # this was a different way.  It's building a dictionary.
                    # if key in dct:
                    #     raise Exception(f"key:{key} is already set and resource existed on the previous loop.")
                    # dct[key] = {
                    #     "value": value,
                    #     "resource": resource,
                    # }
                attrl = attrl.next
            data[bstat_head.name] = dct
            bstat_head = bstat_head.next
    if type(bstat_root) == pbs_ifl.batch_status:
        pbs_ifl.pbs_statfree(bstat_root)
    return data


def autodetect_interface(machine_name=None, debug=False, host=None) -> PBSInterfaceABC:
    try:
        interface = PBSIFLInterface(machine_name=machine_name, host=host, debug=debug)
    except ImportError as err:  # failing to import pbs_ifl
        print(f"failed to import pbs ifl interface, err:{err}")
        try:
            interface = PBSCommandInterface(machine_name=machine_name, debug=debug)
        except ImportError as err:
            print(f"failed to import pbs command interface, err:{err}")
            interface = None
    return interface


def pbs_import_hook_prepare(hook_path: Path, pbs_home: str, callbacks: Optional[list] = None):
    """creates a temporary file in the pbs_home, copies the source file to that file,
    executes callbacks on that file, then imports it."""
    fd, path = tempfile.mkstemp(prefix=hook_path.stem, suffix=hook_path.suffix, dir=pbs_home)
    with os.fdopen(fd, "wb") as f_dst:
        f_dst.write(hook_path.read_bytes())
    hook_path_dst = Path(path)
    hook_file_dst = hook_path_dst.name
    for callback in callbacks:
        callback(hook_path_dst)
    return hook_file_dst


def pbs_import_hook_execute(pbs_interface, hook_name, hook_file, content_encoding, content_type):
    pbs_ifl = pbs_interface.pbs_ifl
    attropl = pbs_interface.get_attropl("input-file", value=hook_file, batch_op=pbs_ifl.SET, next_attropl=None)
    attropl = pbs_interface.get_attropl("content-encoding", value=content_encoding, batch_op=pbs_ifl.SET, next_attropl=attropl)
    attropl = pbs_interface.get_attropl("content-type", value=content_type, batch_op=pbs_ifl.SET, next_attropl=attropl)
    data = pbs_interface.pbs_manager(pbs_ifl.MGR_CMD_IMPORT, pbs_ifl.MGR_OBJ_HOOK, object_name=hook_name, attropl=attropl)
    return data


def pbs_load_enums(pbs=None):
    if pbs is None:
        import pbs
    enum_lookup = {}
    masks = ("REVERSE_NODE_STATE", "REVERSE_ATR_VFLAGS", "REVERSE_HOOK_EVENT")
    for key in (
        "REVERSE_ATR_VFLAGS",
        "REVERSE_BATCH_OPS",
        "REVERSE_BRP_CHOICES",
        "REVERSE_HOOK_EVENT",
        "REVERSE_JOB_STATE",
        "REVERSE_JOB_SUBSTATE",
        "REVERSE_MGR_CMDS",
        "REVERSE_MGR_OBJS",
        "REVERSE_NODE_STATE",
        "REVERSE_RESV_STATE",
    ):
        dct = {}
        mask = "MASK" if key in masks else "LONG"
        for i, value in getattr(pbs, key).items():
            dct[value] = i
        enum_lookup[f"{mask}__{key}"] = dct
    return enum_lookup
