#!/usr/bin/env python3

# Copyright (C) 2024, UChicago Argonne, LLC
# Licensed under the 3-clause BSD license.  See accompanying LICENSE.txt file
# in the top-level directory.

import subprocess
import sys
from pathlib import Path
import shlex

import click
import logging
from dataclasses import dataclass

from PBS_Utils.pbs_util import get_now
from PBS_Utils.pbs_library import PBSIFLInterface, PBSIFLError

logging.basicConfig()
log = logging.getLogger(__name__)


@dataclass
class Settings:
    def __init__(self) -> None:
        self.verbose: int = 0
        # self.restart: bool = False


pass_settings = click.make_pass_decorator(Settings, ensure=True)

# @click.option('-r', '--restart', is_flag=True, show_default=True, default=True, help="restart pbs.")  # fixme move to the top cli


@click.group()
@pass_settings
@click.option("--verbose", "-v", help="Increase verbosity level", count=True)
def cli(settings: Settings, verbose: int) -> None:
    settings.verbose = verbose
    if settings.verbose >= 2:
        log.setLevel(logging.DEBUG)
    elif settings.verbose == 1:
        log.setLevel(logging.INFO)
    else:
        log.setLevel(logging.WARN)


@cli.command()
@pass_settings
@click.option("-c", "--usecp", type=click.STRING, multiple=True)
@click.option(
    "--path",
    type=click.Path(exists=True, file_okay=True, dir_okay=False, path_type=Path),
    default="/var/spool/pbs/mom_priv/config",
)
@click.argument("logevent", type=click.STRING)
def mom_config(settings: Settings, logevent, usecp, path: Path):
    restart = True
    logevent = f"$logevent {logevent}"
    new_cp_dct = {}
    for cp in usecp:
        try:
            nsrc, ndst = cp.split(" ")
        except ValueError:
            print(f"please provide a correct usecp line: hostspec:path_prefix new_prefix")
            sys.exit(1)
        else:
            new_cp_dct[nsrc] = ndst

    line_lst = path.read_text().split("\n")

    new_lines = []
    new_lines.append(logevent)

    old_cp_dct = {}
    for line in line_lst:
        if not line:
            continue
        elif line.startswith("$logevent"):
            continue
        elif line.startswith("$usecp"):
            try:
                ekey, esrc, edst = line.split(" ")
            except ValueError:
                print(f"invalid usecp line found, removing:{line}")
                continue
            else:
                old_cp_dct[esrc] = edst
        else:
            new_lines.append(line)

    for esrc, edst in old_cp_dct.items():
        if esrc not in new_cp_dct:
            new_lines.append(f"$usecp {esrc} {edst}")
        else:
            ndst = new_cp_dct[esrc]
            new_lines.append(f"$usecp {esrc} {ndst}")
            del new_cp_dct[esrc]

    for nsrc, ndst in new_cp_dct.items():
        new_lines.append(f"$usecp {nsrc} {ndst}")

    contents = ("\n".join(new_lines)) + "\n"
    path.write_text(contents)

    #    restart = False
    if restart:
        _restart_pbs(output=True)


# @cli.command()
# @pass_settings
# @click.option('--name', type=click.STRING)
# def mom_configd(settings: Settings, name):
#     pass


def _restart_pbs(output=False):
    cp = subprocess.run(["/etc/init.d/pbs", "restart"], capture_output=True)
    if output:
        print(f"stderr:{cp.stderr.decode('utf-8')}")
        print(f"stdout:{cp.stdout.decode('utf-8')}")
        print(f"return:{cp.returncode}")
    return cp


@cli.command()
@pass_settings
def restart_pbs(settings: Settings):
    cp = _restart_pbs(output=True)
    return cp.returncode


# @cli.result_callback()
# def exit(result, **kwargs):
#     if 'restart' in kwargs:
#         restart()


@cli.command()
@click.option("--mom", type=click.STRING)
@click.option("--name", type=click.STRING)
@click.option("--pstart", type=click.INT)
@click.option("--pend", type=click.INT)
@pass_settings
def create_moms(settings: Settings, mom, name, pstart, pend):
    fast = True
    if fast:
        import pbs_ifl

        pbs_interface = PBSIFLInterface(pbs_ifl=pbs_ifl)
    else:
        pbs_interface = None
    time_now = get_now()
    time_start = time_now
    for i in range(pstart, pend + 1):
        vnode = f"{name}-{i}"
        if fast:
            attropl = pbs_interface.get_attropl("mom", value=mom)
            attropl = pbs_interface.get_attropl("port", value=str(i), next_attropl=attropl)
            # attropl = pbs_interface.get_attropl('state', value='down', next_attropl=attropl)
            data = pbs_interface.pbs_manager(pbs_ifl.MGR_CMD_CREATE, pbs_ifl.MGR_OBJ_NODE, object_name=vnode, attropl=attropl)
            msg = data
        else:
            cmd = f"/opt/pbs/bin/qmgr -c 'create node {vnode} mom={mom},port={i}'"
            cp = subprocess.run(shlex.split(cmd), capture_output=True)
            if cp.returncode != 0:
                print(cp.returncode)
                print(cp.stdout.decode("utf-8"))
                print(cp.stderr.decode("utf-8"))
                break
            msg = cp.stdout.decode("utf-8")
        time_prv = time_now
        time_now = get_now()
        time_delta = (time_now - time_prv).total_seconds()
        print(f"{time_now}Δ({time_delta: >4}):created vnode {vnode} on {mom}: {msg}")
    time_end = get_now()
    time_delta = (time_end - time_start).total_seconds()
    print(f"total Δ({time_delta: >4})")


@cli.command()
@click.option("--mom", type=click.STRING)
@click.option("--name", type=click.STRING)
@click.option("--pstart", type=click.INT)
@click.option("--pend", type=click.INT)
@pass_settings
def delete_moms(settings: Settings, mom, name, pstart, pend):
    fast = True
    if fast:
        import pbs_ifl

        pbs_interface = PBSIFLInterface(pbs_ifl=pbs_ifl)
    else:
        pbs_interface = None
    time_now = get_now()
    time_start = time_now
    for i in range(pstart, pend + 1):
        vnode = f"{name}-{i}"
        if fast:
            try:
                data = pbs_interface.pbs_manager(pbs_ifl.MGR_CMD_DELETE, pbs_ifl.MGR_OBJ_NODE, object_name=vnode)
            except PBSIFLError as err:
                msg = str(err)
            else:
                msg = data
        else:
            cmd = f"/opt/pbs/bin/qmgr -c 'delete node {vnode}'"
            cp = subprocess.run(shlex.split(cmd), capture_output=True)
            if cp.returncode != 0:
                print(cp.returncode)
                print(cp.stdout.decode("utf-8"))
                print(cp.stderr.decode("utf-8"))
                break
            msg = cp.stdout.decode("utf-8")
        time_prv = time_now
        time_now = get_now()
        time_delta = (time_now - time_prv).total_seconds()
        print(f"{time_now}Δ({time_delta: >4}):deleted vnode {vnode} on {mom}: {msg}")
    time_end = get_now()
    time_delta = (time_end - time_start).total_seconds()
    print(f"total Δ({time_delta: >4})")


if __name__ == "__main__":
    cli()
