#!/usr/bin/env python3

# Copyright (C) 2024, UChicago Argonne, LLC
# Licensed under the 3-clause BSD license.  See accompanying LICENSE.txt file
# in the top-level directory.

import io
import sys
import gzip
import click
import logging
from dataclasses import dataclass
from PBS_Utils.pbs_log import process_log_line

try:
    import pbs
except ImportError:
    pbs = None

logging.basicConfig()
log = logging.getLogger(__name__)


@dataclass
class Settings:
    def __init__(self) -> None:
        self.verbose: int = 0


pass_settings = click.make_pass_decorator(Settings, ensure=True)


@click.group()
@pass_settings
@click.option("--verbose", "-v", help="Increase verbosity level", count=True)
def cli(settings: Settings, verbose: int) -> None:
    settings.verbose = verbose
    if settings.verbose > 2:
        log.setLevel(logging.DEBUG)
    elif settings.verbose > 1:
        log.setLevel(logging.INFO)
    else:
        log.setLevel(logging.WARN)


@cli.command()
@pass_settings
@click.option("--pipein", type=click.File("rb"), default=sys.stdin)
@click.option("--pipeout", type=click.File("at"), default=sys.stdout)
def pipe(settings: Settings, pipein, pipeout):
    f = gzip.open(pipein, "rb")
    try:
        line = f.readline()
    except (TypeError, gzip.BadGzipFile):
        f = pipein
        try:
            f.seek(0)  # we read the first two bytes, give them back
        except io.UnsupportedOperation:
            pass  # this was probably piped in.
        line = f.readline()
    while line:
        if type(line) == bytes:
            line = line.decode("utf-8")
        line = process_log_line(line)
        pipeout.write(line)
        # pipeout.flush()
        line = f.readline()


if __name__ == "__main__":
    cli()
