#!/usr/bin/env python3

# Copyright (C) 2024, UChicago Argonne, LLC
# Licensed under the 3-clause BSD license.  See accompanying LICENSE.txt file
# in the top-level directory.

import click
import pbs_ifl
import logging
import pandas as pd
from dataclasses import dataclass
from PBS_Utils.pbs_library import PBSIFLInterface

logging.basicConfig()
log = logging.getLogger(__name__)

pd.set_option("display.max_rows", None)
pd.set_option("display.max_columns", None)
pd.set_option("display.width", None)
pd.set_option("display.max_colwidth", None)
pd.set_option("display.float_format", "{:,.9f}".format)


@dataclass
class Settings:
    def __init__(self) -> None:
        self.verbose: int = 0
        self.server = None


pass_settings = click.make_pass_decorator(Settings, ensure=True)


@click.group(invoke_without_command=True, context_settings={"max_content_width": 132})
@pass_settings
@click.option("--server", help="server host to connect via ifl", default=None)
@click.option("--verbose", "-v", help="Increase verbosity level", count=True)
@click.option("--sortby", multiple=True, help="column(s) to sort by")
@click.option("--resources", "-r", multiple=True)
@click.option(
    "--header",
    show_default=True,
    default="vnode,port,state,njobs,ncpus,jobs,resv,comment",
    help="The header to display, comma separated list.",
)
@click.option("--bare", is_flag=True, show_default=True, default=False, help="no header")
def cli(settings: Settings, server, verbose: int, sortby: list, resources: list, header, bare) -> None:
    settings.server = server
    settings.verbose = verbose
    if settings.verbose >= 2:
        log.setLevel(logging.DEBUG)
    elif settings.verbose == 1:
        log.setLevel(logging.INFO)
    else:
        log.setLevel(logging.WARN)

    header = header.split(",")
    for resource in resources:
        header.append(resource)

    pbs_interface = PBSIFLInterface(host=settings.server, pbs_ifl=pbs_ifl, debug=log.level == logging.DEBUG)
    result = pbs_interface.get_pbs_vnodes()
    vnodes = result["data"]
    node_lst = []
    for vnode_name, dct in vnodes.items():
        node_dct = {}
        node_dct["port"] = dct["Port"] if "Port" in dct else None
        node_dct["comment"] = "" if "comment" not in dct else dct["comment"][:16]
        node_dct["vnode"] = vnode_name
        node_dct["state"] = dct["state"]
        node_dct["resv"] = "" if "resv" not in dct else dct["resv"]
        if "jobs" in dct:
            jobs = dct["jobs"]
            job_lst = jobs.split(",")
            job_lst = sorted(list(set(map(lambda _: _.split(".", 1)[0].strip(), job_lst))))
        else:
            job_lst = []
        node_dct["njobs"] = len(job_lst)
        node_dct["jobs"] = " ".join(job_lst)

        key = "ncpus"
        ncpus_total = int(dct["resources_available"][key]) if key in dct["resources_available"] else 0
        ncpus_free = ncpus_total - int(dct["resources_assigned"][key])
        node_dct[key] = f"{ncpus_free}/{ncpus_total}"

        for key in resources:
            if key in dct["resources_available"]:
                resource_total = int(dct["resources_available"][key])
                resource_free = (
                    resource_total - int(dct["resources_assigned"][key]) if key in dct["resources_assigned"] else resource_total
                )
                node_dct[key] = f"{resource_free}/{resource_total}"
            else:
                node_dct[key] = "0/0"
        node_lst.append(node_dct)

    df = pd.DataFrame(node_lst)
    for col in header:
        missing_cols = []
        if col not in df:
            missing_cols.append(col)
        if not missing_cols:
            df = pd.DataFrame(df, columns=header)
        else:
            log.error(f"header failed, missing columns: {' '.join(missing_cols)}")

    if sortby:
        missing_cols = []
        for col in sortby:
            if col not in df:
                missing_cols.append(col)
        if not missing_cols:
            df = df.sort_values(by=list(sortby))
        else:
            log.error(f"sort failed, missing columns: {' '.join(missing_cols)}")

    if not df.empty:
        print(df.to_string(index=False, header=not bare))
    else:
        print("No nodes found.")


if __name__ == "__main__":
    cli()
