#!/usr/bin/env python3

# Copyright (C) 2024, UChicago Argonne, LLC
# Licensed under the 3-clause BSD license.  See accompanying LICENSE.txt file
# in the top-level directory.

import copy
import datetime
import pbs_ifl
import argparse
import numpy as np
import pandas as pd
from pprint import pformat
from PBS_Utils.pbs_library import PBSIFLInterface
from PBS_Utils.pbs_accounting import cast_to_vnode_list
from PBS_Utils.pbs_util import timedelta_str_to_td, timedelta_to_str


pd.set_option("display.max_rows", None)
pd.set_option("display.max_columns", None)
pd.set_option("display.width", None)
pd.set_option("display.max_colwidth", None)
pd.set_option("display.float_format", "{:,.9f}".format)


def get_argparse():
    # FIXME: switch to click
    parser = argparse.ArgumentParser()
    parser.add_argument("-v", "--debug", dest="debug", action=argparse.BooleanOptionalAction)
    # parser.add_argument('--json', action='store_true')
    parser.add_argument("--history", dest="history", action=argparse.BooleanOptionalAction)
    parser.add_argument("--vnodes", dest="vnodes", action=argparse.BooleanOptionalAction, help="show vnodes.")
    parser.add_argument("--fulljobid", dest="fulljobid", action=argparse.BooleanOptionalAction, help="full jobid.")
    parser.add_argument("--score", dest="score", action=argparse.BooleanOptionalAction)
    parser.add_argument("--comment", dest="comment", action=argparse.BooleanOptionalAction)
    parser.add_argument("-Q", "--queue", dest="queue", action=argparse.BooleanOptionalAction)
    # TODO: header
    # TODO: sortby
    # TODO: add filter for held jobs, etc.
    # TODO: add nodes needed.
    args = parser.parse_args()
    return args


def err_handler(err_str):
    print(err_str, flush=True)


def calculate_remaining(walltime: str, time_used: str):
    td_wall = timedelta_str_to_td(walltime)
    td_used = timedelta_str_to_td(time_used)
    return timedelta_to_str(td_wall - td_used)


def main():
    args = get_argparse()
    debug = True if args.debug else False
    # json_output = True if args.json else False
    fulljobid = True if args.fulljobid else False
    vnodes = True if args.vnodes else False
    history = True if args.history else False
    score = True if args.score else False
    comment = True if args.comment else False
    queue = True if args.queue else False

    pbs_interface = PBSIFLInterface(pbs_ifl=pbs_ifl, debug=debug)

    if queue:
        mode = "queue"
        short_name_lookup = {
            "resources_assigned": "ra",
            "resources_default": "rd",
            "resources_max": "rmax",
            "resources_min": "rmin",
        }
        result_queue = pbs_interface.get_pbs_queues()
        queues_dct = result_queue["data"]
        queue_lst = []
        for queue_name, queue_dct in queues_dct.items():
            new_queue_dct = {}
            new_queue_dct["name"] = queue_name
            for key, value in queue_dct.items():
                if key in short_name_lookup:
                    key = short_name_lookup[key]
                if type(value) == dict:
                    for subkey, subvalue in value.items():
                        new_queue_dct[f"{key}.{subkey}"] = subvalue
                else:
                    new_queue_dct[key] = value
            queue_lst.append(new_queue_dct)
        df = pd.DataFrame(queue_lst)
        df = df.set_index("name")
    else:
        mode = "job"
        columns = [
            "jobid",
            "Job_Name",
            "User",
            "Queued",
            "Walltime",
            "Eligible",
            "Time_Use",
            "Remain",
            "Nodes",
            "S",
            "Account_Name",
            "queue",
        ]
        if score:
            columns.append("score")
            sort_column = "score"
        else:
            sort_column = "jobid"
        if vnodes:
            columns.append("Vnodes")
        if comment:
            columns.append("comment")
        if history:
            result_jobs = pbs_interface.get_pbs_jobs(attrib=None, extend="xt", err_handler=err_handler, score=score)
        else:
            result_jobs = pbs_interface.get_pbs_jobs(attrib=None, extend="t", err_handler=err_handler, score=score)
        jobs = result_jobs["data"]
        job_lst = []
        for jobid, job_dct in jobs.items():
            job_dct = copy.deepcopy(job_dct)
            try:
                stime = job_dct.get("stime", None)
                if not fulljobid and "." in jobid:
                    jobid, _ = jobid.split(".", 1)
                job_dct["jobid"] = jobid
                job_dct["S"] = job_dct["job_state"]
                job_dct["Project"] = job_dct.get("Account_Name", None)
                job_dct["User"], _ = job_dct["Job_Owner"].split("@")
                job_dct["Queued"] = datetime.datetime.utcfromtimestamp(int(job_dct["qtime"]))
                job_dct["Start"] = datetime.datetime.utcfromtimestamp(int(stime)) if stime else None
                job_dct["Walltime"] = job_dct["Resource_List"].get("walltime", None)
                job_dct["Eligible"] = job_dct["eligible_time"]
                # This is odd, resources_used doesn't always exist when a job is running...
                job_dct["Time_Use"] = job_dct.get("resources_used", {}).get("walltime", "00:00:00")  # '   --   '
                if job_dct["Time_Use"] and job_dct["Walltime"] and job_dct["S"] == "R":
                    walltime = job_dct["Walltime"]
                    time_used = job_dct["Time_Use"]
                    job_dct["Remain"] = calculate_remaining(walltime, time_used)
                else:
                    if job_dct["S"] == "E":
                        job_dct["Remain"] = "   --   "
                    else:
                        job_dct["Remain"] = "   --   "
                job_dct["Time_Use"] = job_dct.get("resources_used", {}).get("walltime", "   --   ")
                job_dct["Nodes"] = job_dct["Resource_List"].get("nodect", None)
                if vnodes:
                    vnode_lst = cast_to_vnode_list(job_dct["exec_vnode"]) if job_dct.get("exec_vnode", None) else []
                    job_dct["Vnodes"] = ",".join(vnode_lst)
                job_dct["comment"] = job_dct["comment"] if "comment" in job_dct else ""
                job_lst.append(job_dct)
            except Exception as err:
                print(pformat(job_dct))
                raise
        del jobs
        df = pd.DataFrame(job_lst, columns=columns)
        df = df.set_index("jobid")
        df = df.replace(np.nan, "-", regex=True)
        df = df.sort_values(sort_column, ascending=False)

    if not df.empty:
        # if json_output:
        #     import json
        #     # FIXME: make this work.
        #     print(pformat(df.to_json(date_format='iso')))
        # else:
        print(df)
    else:
        print(f"No {mode}s found.")


if __name__ == "__main__":
    main()
