# Copyright (C) 2024, UChicago Argonne, LLC
# Licensed under the 3-clause BSD license.  See accompanying LICENSE.txt file
# in the top-level directory.

import socket
import pytest
from PBS_Utils.pbs_library import PBSIFLInterface
import os

try:
    import pbs_ifl
except ImportError:
    pbs_ifl = None


@pytest.mark.skipif(pbs_ifl is None, reason="could not import pbs_ifl")
class TestPBSLibrary:
    @classmethod
    def setup_class(cls):
        machine_name = "pbs_generic"
        pbs_host = "pdw-s1"
        debug = True
        cls.pbs_interface = PBSIFLInterface(machine_name=machine_name, host=pbs_host, debug=debug, pbs_ifl=pbs_ifl)

    @classmethod
    def teardown_class(cls):
        cls.pbs_interface.disconnect()

    @pytest.mark.skipif(os.geteuid() != 0, reason="test requires sudo")
    def test_pbs_stathook(self):
        data = self.pbs_interface.pbs_stathook("", None, None)
        print(data)

    def test_pbs_stat_server_00(self):
        server = self.pbs_interface.host
        data = self.pbs_interface.pbs_statserver(server, "python_restart_max_hooks")
        print(data)

    def test_get_pbs_vnodes(self):
        data = self.pbs_interface.get_pbs_vnodes()
        print(data)

    def test_pbs_manager(self):
        self.pbs_interface.get_attropl("python_restart_max_hooks", value="1000000000")
        data = self.pbs_interface.pbs_manager("set", "server")
        print(data)

    def test_qstat(self):
        data = self.pbs_interface.get_pbs_server()
        print(data)

    def test_get_pbs_jobs(self):
        """Assuming no jobs have been started (clean system)"""
        data = self.pbs_interface.get_pbs_jobs()
        assert len(data["data"]) == 0
        assert data["metadata"]["machine_name"] == "pbs_generic"

    def test_get_pbs_nodes(self):
        """Assuming 4 node system"""
        data = self.pbs_interface.get_pbs_nodes()
        assert len(data["data"]) == 4
        assert data["metadata"]["machine_name"] == "pbs_generic"

    def test_get_pbs_reservations(self):
        """Assuming no reservations present on system (clean system)"""
        data = self.pbs_interface.get_pbs_reservations()
        assert len(data["data"]) == 0
        assert data["metadata"]["machine_name"] == "pbs_generic"

    @pytest.mark.skipif(os.geteuid() == 0, reason="test requires a valid user, not 0")
    def test_pbs_submit(self):
        """Submitting job to pbs_ifl and checking"""
        assert len(self.pbs_interface.get_pbs_jobs()["data"]) == 0

        job_script = "tests/submit_script.sh"
        attropl_list = [{"name": "Resource_List", "value": "1:ncpus=1", "resource": "select"}]
        jobid = self.pbs_interface.pbs_submit(job_script, attropl_list)
        assert len(jobid) > 1

        assert len(self.pbs_interface.get_pbs_jobs()["data"]) == 1

        # delete job and test
        self.pbs_interface.pbs_delete(jobid, wait=True)

        # wait for job to disappear
        assert len(self.pbs_interface.get_pbs_jobs()["data"]) == 0

        # check for 0
        data = self.pbs_interface.get_pbs_jobs()
        assert len(data["data"]) == 0

    @pytest.mark.skipif(os.geteuid() == 0, reason="test requires a valid user, not 0")
    def test_pbs_submit2(self):
        """Submitting job to pbs_ifl and checking job data (also assuming clean)"""
        assert len(self.pbs_interface.get_pbs_jobs()["data"]) == 0

        data = self.pbs_interface.get_pbs_jobs()
        assert len(data["data"]) == 0
        assert data["metadata"]["machine_name"] == "pbs_generic"

        # submitting job
        job_script = "tests/submit_script.sh"
        attropl_list = [{"name": "Resource_List", "value": "1:ncpus=1", "resource": "select"}]
        joblist = []
        for i in range(2):
            jobid = self.pbs_interface.pbs_submit(job_script, attropl_list)
            assert len(jobid) > 0
            joblist.append(jobid)

        # recheck number of jobs
        data = self.pbs_interface.get_pbs_jobs()
        assert len(data["data"]) == 2
        assert data["metadata"]["machine_name"] == "pbs_generic"
        for job in data["data"].keys():
            assert data["data"][job]["Resource_List"]["ncpus"] == "1"

        # kill the jobs
        for jobid in joblist:
            self.pbs_interface.pbs_delete(jobid, wait=True)

        # make sure all jobs are gone
        assert len(self.pbs_interface.get_pbs_jobs()["data"]) == 0

    @pytest.mark.skipif(pbs_ifl is None, reason="could not import pbs_ifl")
    def test_get_pbs_queues_00(self):
        result_dct = self.pbs_interface.get_pbs_queues()
        queues_dct = result_dct["data"]
        assert len(queues_dct) >= 0

    @pytest.mark.skipif(pbs_ifl is None, reason="could not import pbs_ifl")
    def test_get_pbs_server_00(self):
        server = self.pbs_interface.get_pbs_server_host()
        assert server == socket.gethostname()

        result_dct = self.pbs_interface.get_pbs_server()
        server_dct = result_dct["data"]
        assert len(server_dct) >= 0

        attrl = self.pbs_interface.get_attrl("python_restart_max_hooks")
        result_dct = self.pbs_interface.get_pbs_server(attrl=attrl)
        server_dct = result_dct["data"][server]
        assert len(server_dct) == 1

        attrl_mh = self.pbs_interface.get_attrl("python_restart_max_hooks")
        attrl_mo = self.pbs_interface.get_attrl("python_restart_max_objects", next_attrl=attrl_mh)
        result_dct = self.pbs_interface.get_pbs_server(attrl=attrl_mo)
        server_dct = result_dct["data"][server]
        assert len(server_dct) == 2
